import java.util.LinkedList;
import java.util.Random;


public class Environment 
{
	int fitness;
	
	int maxOperandInterval = 100;
	int maxFors = 3;
	int maxIfs = 3;
	
	int fors;
	int ifs;
	
	Function firstFunction;
	LinkedList<Integer> integerVariables;
	LinkedList<int[]> integerArrays;
	LinkedList<Integer> forLoopIndices = new LinkedList<Integer>();
	
	LinkedList<Integer> integerVariablesBackup= new LinkedList<Integer>();
	LinkedList<int[]> integerArraysBackup= new LinkedList<int[]>();
	LinkedList<Integer> forLoopIndicesBackup = new LinkedList<Integer>();
	
	int forMinIndex, forMaxIndex, maxFun;
	
	public Environment(LinkedList<Integer> variables, LinkedList<int[]> arrays, int forMinIndex, int forMaxIndex, int maxFun)
	{
		fors = 0;
		ifs = 0;
		
		this.integerVariables = variables;
		this.integerArrays = arrays;
		
		for(int var : integerVariables)
			integerVariablesBackup.add(var);
		for(int[] var : integerArrays)
			integerArraysBackup.add(var);
		
		this.forMaxIndex = forMaxIndex;
		this.forMinIndex = forMinIndex;
		this.maxFun = maxFun;
		
		firstFunction = generateProgram(maxFun, 0);
		
		for(int var : forLoopIndices)
			forLoopIndicesBackup.add(var);
	}
	
	public int getIntegerVariableAt(int index)
	{
		return integerVariables.get(index);
	}
	
	public int[] getIntegerArrayAt(int index)
	{
		return integerArrays.get(index);
	}
	
	public int getLoopIndiceAt(int index)
	{
		return forLoopIndices.get(index);
	}
	
	public void setLoopIndice(int index, int value)
	{
		forLoopIndices.set(index, value);
	}
	
	public void setOwnLoopIndice(int value)
	{
		if(forLoopIndices.size() == 0)
			forLoopIndices.add(value);
		else
			forLoopIndices.set(forLoopIndices.size()-1, value);
	}
	
	public void setIntegerVariableAt(int index, int variable)
	{
		integerVariables.set(index, variable);
	}
	
	public void runProgram()
	{
		firstFunction.execute();
	}
	
	public void resetProgram()
	{
		integerVariables.clear();
		integerArrays.clear();
		forLoopIndices.clear();
		
		for(int var : integerVariablesBackup)
			integerVariables.add(var);
		for(int[] var : integerArraysBackup)
			integerArrays.add(var);
		for(int var : forLoopIndicesBackup)
			forLoopIndices.add(var);
	}
	
	public Function generateProgram(int maxFun, int forLoopDepth)
	{
		Random random = new Random();
		int funNumber = random.nextInt(maxFun) + 1;
		Function beginning;
		
		Function current = beginning = generateFunction(random, forLoopDepth);
		
		for(int i=0; i<funNumber; i++)
		{
			Function newFunction = generateFunction(random,forLoopDepth);
			current.setNeighbour(newFunction);
			current = newFunction;
		}
		
		return beginning;
	}
	
	 Function generateFunction(Random random, int forLoopDepth)
	{
		int randomFunction = random.nextInt(8);
		while(fors == maxFors && randomFunction == 0 || ifs == maxIfs && (randomFunction==2 || randomFunction ==3))
			randomFunction = random.nextInt(6);
			
		Function function = null;
		
		switch(randomFunction)
		{
			case 0: //for-loop
				fors++;
				function = generateForLoop(random, forMinIndex, forMaxIndex, forLoopDepth);
				break;
			case 1: //swap
				function = generateSwap(random);
				break;
			case 2: //ifSmaller
				ifs++;
				function = generateIfSmaller(random, forLoopDepth);
				break;
			case 3: //ifGreater
				ifs++;
				function = generateIfGreater(random, forLoopDepth);
				break;
			case 4: //add
				function = generateAdd(random);
				break;
			case 5: //multiply
				function = generateMultiply(random);
				break;
			case 6: //divide
				function = generateDivide(random);
				break;
			case 7:
				function = generateForLoopWithOperation(random, forMinIndex, forMaxIndex, forLoopDepth);
				break;
		}
		
		return function;
	}
	
	private Function generateDivide(Random random)
	{
		int index = random.nextInt(integerVariables.size());
		int value = random.nextInt(maxOperandInterval) - maxOperandInterval / 2;
		if(value == 0)
			value ++;
		return new MultiplyInteger(index, value, this);
	}
	
	private Function generateMultiply(Random random)
	{
		int index = random.nextInt(integerVariables.size());
		int value = random.nextInt(maxOperandInterval) - maxOperandInterval / 2;
		return new MultiplyInteger(index, value, this);
	}
	
	private Function generateAdd(Random random)
	{
		int index = random.nextInt(integerVariables.size());
		int value = random.nextInt(maxOperandInterval) - maxOperandInterval / 2;
		return new AddInteger(index, value, this);
	}
	
	private Function generateIfGreater(Random random, int forLoopDepth)
	{		
		IfGreater function = null;
		int variableOrIndex = random.nextInt(2);
		if(variableOrIndex == 0) //index
		{
			int first = random.nextInt(integerArrays.get(integerArrays.size()-1).length);
			int second = random.nextInt(integerArrays.get(integerArrays.size()-1).length);
			function = new IfGreater(0, first, second, this, integerArrays.size()-1);
		}
		else
		{
			int first = random.nextInt(integerVariables.size());
			int second = random.nextInt(integerVariables.size());
			function = new IfGreater(1, first, second, this, -1);
		}
		
		function.setTFNeighbours(generateFunction(random, forLoopDepth), generateFunction(random, forLoopDepth));
		
		return function;
	}

	private Function generateIfSmaller(Random random, int forLoopDepth)
	{
		IfSmaller function = null;
		int variableOrIndex = random.nextInt(2);
		if(variableOrIndex == 0) //index
		{
			int first = random.nextInt(integerArrays.get(integerArrays.size()-1).length);
			int second = random.nextInt(integerArrays.get(integerArrays.size()-1).length);
			function = new IfSmaller(0, first, second, this, integerArrays.size()-1);
		}
		else
		{
			int first = random.nextInt(integerVariables.size());
			int second = random.nextInt(integerVariables.size());
			function = new IfSmaller(1, first, second, this, -1);
		}
		
		function.setTFNeighbours(generateFunction(random, forLoopDepth), generateFunction(random, forLoopDepth));
		
		return function;
	}
	
	private Function generateSwap(Random random) 
	{
		Function function = null;
		int variableOrIndex = random.nextInt(2);
		if(variableOrIndex == 0 && forLoopIndices.size()>1) //index
		{
			int index1 = random.nextInt(forLoopIndices.size());
			int index2 = random.nextInt(forLoopIndices.size());
			while(index2 == index1 && forLoopIndices.size()>1)
				index2 = random.nextInt(forLoopIndices.size());
			
			function = new Swap(0, index1, index2, integerArrays.size()-1, this);
		}
		else
		{
			int index1 = random.nextInt(integerVariables.size());
			int index2 = random.nextInt(integerVariables.size());
		/*	while(index2 == index1) 
				index2 = random.nextInt(integerVariables.size());
			*/
			function = new Swap(1, index1, index2, integerArrays.size()-1, this);
		}
		return function;
	}
	
	private ForLoop generateForLoop(Random random, int minIndex, int maxIndex, int depth)
	{
		int start;
		if(random.nextInt(1) == 0 || forLoopIndices.size()==0)
		{
			start = random.nextInt(maxIndex-minIndex) + minIndex;
		}
		else
		{
			start = forLoopIndices.get(random.nextInt(forLoopIndices.size()));
		}
		
		int end;
		if(random.nextInt(1) == 0)
		{
			end = random.nextInt(maxIndex-minIndex) + minIndex;
		}
		else
		{
			end = integerArrays.getLast().length-1;
		}
		
		forLoopIndices.add(start);
		Function body = generateProgram(maxFun, depth+1);
		ForLoop loop = new ForLoop(depth, start, end, body, this);
		return loop;
	}
	
	private ForLoopWithOperation generateForLoopWithOperation(Random random, int minIndex, int maxIndex, int depth)
	{
		int start;
		if(random.nextInt(1) == 0 || forLoopIndices.size()==0)
		{
			start = random.nextInt(maxIndex-minIndex) + minIndex;
		}
		else
		{
			start = forLoopIndices.get(random.nextInt(forLoopIndices.size()));
		}
		
		int end;
		if(random.nextInt(1) == 0)
		{
			end = random.nextInt(maxIndex-minIndex) + minIndex;
		}
		else
		{
			end = integerArrays.getLast().length-1;
		}
		
		forLoopIndices.add(start);
		Function body = generateProgram(maxFun, depth+1);
		ForLoopWithOperation loop = new ForLoopWithOperation(depth, start, end, body, this, random.nextInt(3), (random.nextInt(maxOperandInterval)) );
		return loop;
	}

}
