import java.util.LinkedList;
import java.util.Random;

public class ForLoop extends Function 
{
	int depth;
	int start;
	int end;
	
	Function body;
	LinkedList<ForLoop> parentLoops;
	int currentIndex;
	
	public ForLoop(int depth, int start, int end, Function body, Environment environment)
	{
		this.depth = depth;
		this.start = start;
		this.end = end;
		
		parameters.add(depth);
		parameters.add(start);
		parameters.add(end);
		this.body = body;
		this.environment = environment;
	}
	
	public void setBody(Function body)
	{
		this.body = body;
	}
	
	public void setParentLoops(LinkedList<ForLoop> parents)
	{
		this.parentLoops = parents;
	}
	
	public void setNeighbour(Function neighbour)
	{
		neighbours.clear();
		if(neighbour != null)
			neighbours.add(neighbour);
	}

	public int getIndex()
	{
		return currentIndex;
	}
	
	public int getDepth()
	{
		return (Integer)parameters.get(0);
	}
	
	@Override
	public void execute() 
	{
		try {
			for(currentIndex = (Integer) parameters.get(1); currentIndex <=  (Integer) parameters.get(2); currentIndex ++)
			{
				environment.setOwnLoopIndice(currentIndex);
				body.execute();
			}
			
			if(neighbours.size() > 0)
				neighbours.get(0).execute();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	public Function deepCopy(Environment newEnvironment) 
	{
		Function copy;
		Function firstFunction = new ForLoop(depth, start, end, this.body.deepCopy(newEnvironment), newEnvironment);
		Function current = this;
		copy = firstFunction;
		
		while(current.getNextFunction() != null && firstFunction.getNextFunction() != null)
		{
			firstFunction.setNeighbour(current.getNextFunction().deepCopy(newEnvironment));
			current = current.getNextFunction();
			firstFunction = firstFunction.getNextFunction();
		}
		
		return copy;
	}

	@Override
	public void mutateParameter(Random random, int interval) 
	{
		if(random.nextInt() == 0)
		{
			start = start + random.nextInt(interval) - interval/2;
			parameters.set(1, start);
		}
		else
		{
			end = end + random.nextInt(interval) - interval/2;
			parameters.set(2, end);
		}
			
	}

	
	
}
