import java.util.Arrays;
import java.util.LinkedList;
import java.util.Random;


public class GeneticOperators 
{
	public static int sortingFitness(Environment individual)
	{
		int fitness = 0;
		Random random = new Random();
		int array[] = {1,2,3,4,5};
		
		for(int i=0; i<100; i++)
		{
			individual.integerArrays.clear();
			int[] permutated = getRandomPermutation(array, random);
			individual.integerArrays.add(permutated);
			individual.runProgram();
			
			int[] finalArray = individual.getIntegerArrayAt(0);
			
			Arrays.sort(array);
			
			for(int j=0; j<finalArray.length; j++)
				if(finalArray[j] == array[j])
					fitness++;
			
			individual.resetProgram();
		}
		 
		LinkedList<String> lines = Loops.printProgram(new LinkedList<String>(), individual.firstFunction, 0);
		
		return fitness ;
	}
	
	public static int[] getRandomPermutation (int [] array, Random random)
	{

	    for(int i = 0; i < array.length; i++)
	    {
	        int ran = i + random.nextInt (array.length-i);
	        // perform swap
	        int temp = array[i];
	        array[i] = array[ran];
	        array[ran] = temp;
	    }                       
	    return array;
	}
	
	public static Environment crossover(Environment parent1, Environment parent2, Random random)
	{
		LinkedList<Integer> integerVariables = new LinkedList<Integer>();
		LinkedList<int[]> integerArrays = new LinkedList<int[]>();
		LinkedList<Integer> forLoopIndices = new LinkedList<Integer>();
		
		for(int i=0; i < parent1.integerVariables.size(); i++)
		{
			if(random.nextInt(2)==0)
				integerVariables.add(parent1.getIntegerVariableAt(i));
			else
				integerVariables.add(parent2.getIntegerVariableAt(i));
		}
		
		for(int i=0; i < parent1.integerArrays.size(); i++)
		{
			if(random.nextInt(2)==0)
				integerArrays.add(parent1.getIntegerArrayAt(i));
			else
				integerArrays.add(parent2.getIntegerArrayAt(i));
		}
		
		Environment newEnvironment = new Environment(integerVariables, integerArrays, parent1.forMinIndex, parent1.forMaxIndex, parent1.maxFun);
		newEnvironment.forLoopIndices = forLoopIndices;
		
		Function fun1 = parent1.firstFunction;
		Function fun2 = parent2.firstFunction;
		Function first;
		
		if(random.nextInt(2) == 0)
		{
			first = fun1.deepCopy(newEnvironment);
			fun1 = fun1.getNextFunction();
		}
		else
		{
			first = fun2.deepCopy(newEnvironment);
			fun2 = fun2.getNextFunction();
		}
		
		Function current = first;
		
		newEnvironment.firstFunction = first;
		
		float nextFunctionProbability = 1;
		
		while(current != null)
		{
			if(random.nextFloat() > nextFunctionProbability)
				break;
			
			if(random.nextInt(2) == 0 && fun1 != null)
			{
				current.setNeighbour(fun1.deepCopy(newEnvironment));
				
				if(fun1 instanceof ForLoop)
				{
					newEnvironment.forLoopIndices.add((Integer)fun1.parameters.get(1));
					
					if(random.nextInt(2) == 0)
						fun1 = fun1.getNextFunction();
					else
						fun1 = ((ForLoop) fun1).body;
				}
				else
					fun1 = fun1.getNextFunction();
					
			}
			else if( fun2 != null)
			{
				current.setNeighbour(fun2.deepCopy(newEnvironment));
				
				if(fun2 instanceof ForLoop)
				{
					newEnvironment.forLoopIndices.add((Integer)fun2.parameters.get(1));
					if(random.nextInt(2) == 0)
						fun2 = fun2.getNextFunction();
					else
						fun2 = ((ForLoop) fun2).body;
				}
				else
					fun2 = fun2.getNextFunction();
			}
			
			current = current.getNextFunction();
			nextFunctionProbability /= 2;
		}
		
		return newEnvironment;
	}
	
	public static void mutationByDeletion(float mutationFactor, Environment individual, Random random)
	{
		Function fun = individual.firstFunction;
		Function funNeighbour = fun.getNextFunction();
		
		if(random.nextFloat() <= mutationFactor)
		{
			while(random.nextInt(2)==0 && funNeighbour.getNextFunction() != null)
			{
				
				if(funNeighbour instanceof ForLoop && random.nextInt(2)==0)
				{
					fun = ((ForLoop) funNeighbour).body;
					if(fun instanceof ForLoop)
						funNeighbour = ((ForLoop) fun).body;
					else
						funNeighbour = fun.getNextFunction();
					
					while(random.nextInt(2)==0 && funNeighbour != null)
					{
						fun = funNeighbour;
						if(funNeighbour instanceof ForLoop)
							funNeighbour = ((ForLoop) funNeighbour).body;
						else
							funNeighbour = funNeighbour.getNextFunction();
					}
				}
				
				else
				{	
					fun = funNeighbour;
					funNeighbour = funNeighbour.getNextFunction();
				}
			}
			
			if(funNeighbour!=null)
			{
					fun.setNeighbour(funNeighbour.getNextFunction());
					funNeighbour.setNeighbour(null);
			}
			else
			{
				if(fun.equals(individual.firstFunction))
					individual.firstFunction = fun.getNextFunction();
				else
				fun = null;
			}
		}
	}
	
	public static void mutationOfParameters(float mutationFactor, Environment individual, Random random, int interval)
	{
		Function current = individual.firstFunction;
		
		if(random.nextFloat() <= mutationFactor)
		{
			while(random.nextInt(1)==0)
			{
				if(current instanceof ForLoop)
					current = ((ForLoop) current).body;
				else if(current instanceof IfSmaller || current instanceof IfGreater)
					current = current.neighbours.get(random.nextInt(2));
				else if (current.getNextFunction() != null)
					current = current.getNextFunction();
				else break;
			}
			
			current.mutateParameter(random, interval);
		}
	}
	
	public static LinkedList<Environment> tournament(LinkedList<Environment> population, int tournamentSize, Random random)
	{
		LinkedList<Environment> tournamentResults = new LinkedList<Environment>();
		LinkedList<Environment> tournament = new LinkedList<Environment>();
		
		for(int i=0; i < tournamentSize; i++)
		{
			int index = random.nextInt(population.size());
			while( (tournament.contains(population.get(index))))
				index = random.nextInt(population.size());
			tournament.add(population.get(index));
		}
		
		for(Environment individual : tournament)
			individual.fitness = GeneticOperators.sortingFitness(individual);
				
		int indexOfFirst = 0;
		int maxFitness = tournament.get(0).fitness;
		
		for(int i=1; i < tournamentSize; i++)
		{
			if(tournament.get(i).fitness > maxFitness)
			{
				indexOfFirst = i;
				maxFitness = tournament.get(i).fitness;
			}
		}
		
		tournamentResults.add(tournament.get(indexOfFirst));
		tournament.remove(indexOfFirst);
		
		int indexOfSecond = 0;
		maxFitness = tournament.get(0).fitness;
		
		for(int i=1; i < tournament.size(); i++)
		{
			if(tournament.get(i).fitness > maxFitness)
			{
				indexOfSecond = i;
				maxFitness = tournament.get(i).fitness;
			}
		}
		
		tournamentResults.add(tournament.get(indexOfSecond));
		tournament.remove(indexOfSecond);
		
		int indexOfThird = 0;
		int minFitness = tournament.get(0).fitness;
		
		for(int i=1; i < tournament.size(); i++)
		{
			if(tournament.get(i).fitness < minFitness)
			{
				indexOfThird = i;
				minFitness = tournament.get(i).fitness;
			}
		}
		
		tournamentResults.add(tournament.get(indexOfThird));
		
		return tournamentResults;
	}
}
