import java.util.Random;


public class Swap extends Function 
{	
	int indicesOrVariables, index1, index2, array;
	
	public Swap(int indicesOrVariables, int index1, int index2, int array, Environment environment)
	{
		this.indicesOrVariables = indicesOrVariables;
		this.index1 = index1;
		this.index2 = index2;
		this.array = array;
		this.environment = environment;
		
		if(environment.forLoopIndices.size() < 2)
		{
			if(index1 >= environment.integerVariables.size())
				index1 = environment.integerVariables.size()-1;
			if(index2 >= environment.integerVariables.size())
				index2 = environment.integerVariables.size()-1;
			
			parameters.add(environment.getIntegerVariableAt(index1));
			parameters.add(environment.getIntegerVariableAt(index2));
		}
		else
			switch(indicesOrVariables)
			{
				case 0: //loop indices
					if(index1 >= environment.forLoopIndices.size())
						index1 = environment.forLoopIndices.size()-1;
					if(index2 >= environment.forLoopIndices.size())
						index2 = environment.forLoopIndices.size()-1;
					parameters.add(environment.getLoopIndiceAt(index1));
					parameters.add(environment.getLoopIndiceAt(index2));
					break;
				case 1: //variables
					if(index1 >= environment.integerVariables.size())
						index1 = environment.integerVariables.size()-1;
					if(index2 >= environment.integerVariables.size())
						index2 = environment.integerVariables.size()-1;
					parameters.add(environment.getIntegerVariableAt(index1));
					parameters.add(environment.getIntegerVariableAt(index2));
					break;
			}
		parameters.add(environment.getIntegerArrayAt(array));
	}
	
	public void setNeighbour(Function neighbour)
	{
		neighbours.clear();
		if(neighbour != null)
			neighbours.add(neighbour);
	}
	
	@Override
	public void execute()
	{
		int[] array = (int[]) parameters.get(2);
		if(array.length - 1 < (Integer) parameters.get(0) || array.length - 1 < (Integer) parameters.get(1))
			return;
		
		try
		{
			int temp = ((int[]) parameters.get(2)) [(Integer) parameters.get(0)];
			((int[]) parameters.get(2)) [(Integer) parameters.get(0)] = ((int[]) parameters.get(2)) [(Integer) parameters.get(1)];
			((int[]) parameters.get(2)) [(Integer) parameters.get(1)] = temp;
			
			if(neighbours.size() > 0)
				neighbours.get(0).execute();
		}
		catch(Exception e)
		{
			e.printStackTrace();
			//throw new LoopsException("Swap function not working: " + e.getMessage());
		}
	}

	@Override
	public Function deepCopy(Environment newEnvironment) 
	{
		Function copy;
		Function firstFunction = new Swap(indicesOrVariables, index1, index2, array, newEnvironment);
		Function current = this;
		copy = firstFunction;
		
		while(current.getNextFunction() != null  && firstFunction.getNextFunction() != null)
		{
			firstFunction.setNeighbour(current.getNextFunction().deepCopy(newEnvironment));
			current = current.getNextFunction();
			firstFunction = firstFunction.getNextFunction();
		}
		
		return copy;
	}

	@Override
	public void mutateParameter(Random random, int interval) 
	{
		if(random.nextInt() == 0)
		{
			index1 += random.nextInt(interval) - interval/2;
			if(index1 < 0)
				index1 = 0;
			parameters.set(0, index1);
		}
		else
		{
			index2 += random.nextInt(interval) - interval/2;
			if(index2 < 0)
				index2 = 0;
			parameters.set(1, index2);
		}
	}

}
